/*
 AnaGram, a System for Syntax Directed Programming

 Character Sink Class Definitions

 Copyright (c) 1993, Parsifal Software.
 All Rights Reserved.
*/

#ifndef CHARSINK_H
#define CHARSINK_H

#include <stdio.h>
#include "assert.h"
#include <stack.h>


// Class Definition for character_sink

class character_sink {
public:
  virtual ~character_sink() {}
  virtual character_sink& operator << (int c) = 0;
  virtual character_sink& operator << (char *str) = 0;
  virtual character_sink& operator << (unsigned char *str) = 0;
  virtual character_sink& printf(char *, ...) = 0;
};


// Class Definition for output_file

class output_file : public character_sink {
  FILE *file;
  char *name;
  int error_flag;
  int copy_flag;                            // is it live or memorex?
public:


// Constructors

  output_file() {
    file = stdout;
    error_flag = copy_flag = 0;
  }

  output_file(char *path) {
    name = path;
    if (name == NULL) file = stdout;
    else file = fopen(name = path, "wt");
    if(file == NULL) file = stdout;
    error_flag = copy_flag = 0;
  }

  output_file(output_file &f) {
    *this = f;
    copy_flag++;
  }


// Destructor

  ~output_file() {if (copy_flag == 0) fclose(file);}


// Xmit data to sink

  character_sink& operator << (int c) {
    error_flag = putc(c,file);
    return *this;
  }

  character_sink& operator << (char *str) {
    fputs(str,file);
    return *this;
  }

  character_sink& operator << (unsigned char *str) {
    fputs((char *)str, file);
    return *this;
  }

  character_sink& printf(char *, ...);


// Check for error

  friend int error(output_file &f);
};

inline int error(output_file &f) {
  return f.error_flag < 0;
}

class string_accumulator : public character_sink {
  char *cs;                // character storage
  unsigned *xs;            // index storage
  unsigned csx;            // next available character
  unsigned xsx;            // next available index
  unsigned csmax;          // max number of characters
  unsigned xsmax;          // max number of indices
  int copy_flag;           // is it live or is it memorex?
public:


// Constructor

  string_accumulator(int nc, int nx = 1);
  string_accumulator(string_accumulator &);


// Destructor

  ~string_accumulator();


// Reset string_accumulator

  friend string_accumulator &reset(string_accumulator &);


// Change stack level

  string_accumulator& operator ++(){        // increment stack level
    assert(copy_flag == 0);
    assert(xsx);
    xs[--xsx] = csx;
    return *this;
  }

  string_accumulator operator ++(int x){    // increment stack level
    string_accumulator cpy(*this);          // make copy for return value
    assert(copy_flag == 0);
    assert(xsx);
    xs[--xsx] = csx;
    return cpy;
  }

  string_accumulator& operator --() {       // decrement stack level
    assert(copy_flag == 0);
    csx = xs[xsx++];                        // discard top string
    assert(xsx < xsmax);
    return *this;
  }

  string_accumulator operator --(int x) {   // post decrement
    string_accumulator cpy(*this);          // make copy for return value
    assert(copy_flag == 0);
    csx = xs[xsx++];
    assert(xsx < xsmax);
    return cpy;
  }


// Append data to stack

  character_sink& operator << (int c) {     //append char to top string
    assert(copy_flag == 0);
    assert(csx < csmax);
    cs[csx++] = c;
    return *this;
  }

  character_sink& operator << (char *str);  // append string to top string

  character_sink& operator << (unsigned char *str);

  character_sink &operator << (string_accumulator &s) {
    assert(this != &s);                     // prevent disaster
    return *this << (char *) s;
  }

  character_sink& printf(char *, ...);


// Pop character from stack

  string_accumulator &operator >> (int &c); // pop last character


// Access character on stack

  char &operator [](unsigned);                   // inspect character


// Retrieve pointer to top string

  operator char *() {                       // pointer to top string
    cs[csx] = 0;
    return &cs[xs[xsx]];
  }

  operator unsigned char *() {              // pointer to top string
    cs[csx] = 0;
    return (unsigned char *) &cs[xs[xsx]];
  }


// Concatenate strings

  friend string_accumulator& concat(string_accumulator &);


// Make permanent copy of string

  friend char *copy(string_accumulator &);


// Get length of string

  friend unsigned size(string_accumulator &s);

};

inline unsigned size(string_accumulator &s) {
  assert(s.xsx < s.xsmax);
  return s.csx - s.xs[s.xsx];
}

#endif
